@@ -1,22 +1,31 @@
 # -*- coding: utf-8 -*-
 """BAT ALGORITHM"""
 
-"""
-Yang, Xin‐She, and Amir Hossein Gandomi. 
-Bat algorithm: a novel approach for global engineering optimization. 
-Engineering computations (2012). 
-https://arxiv.org/pdf/1211.6663.pdf
 
-Yang, Xin‐She. Nature-inspired optimization algorithms (2021).
-
-loudness and pulse rate generated for each bat (not same for all) (initial*2*rand).
-"""
 
 import numpy as np
 from ._optimizer import Optimizer, CandidateState 
 
 
 class Bat(CandidateState):
+    """BA Bat class. A Bat is a member of a BA swarm.
+    
+    Parameters
+    ----------
+    V : ndarray
+        Bat velocity.
+    Freq : ndarray
+        Bat frequency.    
+    A : float
+        Bat loudness.
+    r : float
+        Bat pulse rate.
+         
+    Returns
+    -------
+    agent : Bat
+        Bat instance.
+    """
     
     def __init__(self, optimizer: Optimizer):
         CandidateState.__init__(self, optimizer)
@@ -28,20 +37,45 @@ class Bat(CandidateState):
 
 
 class BA(Optimizer):
-    """Bat Algorithm class"""
+    """Bat Algorithm method class.
+    
+    References: [1] Yang, Xin‐She, and Amir Hossein Gandomi. Bat algorithm: a novel approach 
+    for global engineering optimization. Engineering computations (2012). https://arxiv.org/pdf/1211.6663.pdf, 
+    [2] Yang, Xin‐She. Nature-inspired optimization algorithms (2021).
+    
+    In this implementation loudness **A** and pulse rate **r** are generated for each Bat seperately (initial*2*rand).
+    
+    Parameters
+    ----------
+    variant : str
+        Name of the BA variant. Default: ``Vanilla``.
+    params : dict
+        A dictionary of BA parameters.
+    """
 
     def __init__(self):
-        """Initialization"""
         Optimizer.__init__(self)
 
-        self.method = 'Vanilla'
+        self.variant = 'Vanilla'
         self.params = {}
 
     def _check_params(self):
+        """Private method which performs some BA-specific parameter checks
+        and prepares the parameters to be validated by Optimizer._check_params.
+
+        Returns
+        -------
+        None
+            Nothing
+        """
+        
         defined_params = list(self.params.keys())
         mandatory_params, optional_params = [], []
+        
+        if 'bat_swarm_size' in self.params:
+            self.params['bat_swarm_size'] = int(self.params['bat_swarm_size'])
 
-        if self.method == 'Vanilla':
+        if self.variant == 'Vanilla':
             mandatory_params = 'bat_swarm_size loudness pulse_rate alpha gamma freq_range'.split()
 
             if 'bat_swarm_size' not in self.params:
@@ -63,32 +97,37 @@ class BA(Optimizer):
                 self.params['freq_range'] = [0, 1]
                 defined_params += 'freq_range'.split()
         else:
-            assert False, f'Unknown method! {self.method}'
+            assert False, f'Unknown variant! {self.variant}'
 
         Optimizer._check_params(self, mandatory_params, optional_params, defined_params)
 
 
     def _init_method(self):
-        
-        err_msg = None
+        """Private method for initializing the BA optimizer instance.
+        Initializes and evaluates the swarm.
 
-        # Bounds for position and velocity
-        self.lb = np.array(self.lb)
-        self.ub = np.array(self.ub)
+        Returns
+        -------
+        err_msg : str or None
+            Error message (if any).
+        """
+
+        self._evaluate_initial_candidates()
+        err_msg = None
 
         # Generate a swarm
         self.cS = np.array([Bat(self) for c in range(self.params['bat_swarm_size'])], dtype=Bat)
 
         # Generate initial positions
+        n0 = 0 if self._cs0 is None else self._cs0.size
         for p in range(self.params['bat_swarm_size']):
             
             # Random position
             self.cS[p].X = np.random.uniform(self.lb, self.ub)
             
             # Using specified particles initial positions
-            if self.X0 is not None:
-                if p < np.shape(self.X0)[0]:
-                    self.cS[p].X = self.X0[p]
+            if p < n0:
+                self.cS[p] = self._cs0[p].copy()
             
             # Generate velocity
             self.cS[p].V = 0.0
@@ -102,25 +141,35 @@ class BA(Optimizer):
             # Pulse rate
             self.cS[p].r = self.params['pulse_rate']*2*np.random.uniform()
 
-        # Evaluate        
-        self.collective_evaluation(self.cS)
-        # if all candidates are NaNs       
+        # Evaluate
+        if n0 < self.params['bat_swarm_size']:
+            err_msg = self.collective_evaluation(self.cS[n0:])
+
+        # if all candidates are NaNs
         if np.isnan([cP.f for cP in self.cS]).all():
             err_msg = 'ALL CANDIDATES FAILED TO EVALUATE.'
         if err_msg:
             return err_msg
         
-        self.cB = np.array([cP.copy() for cP in self.cS], dtype=CandidateState)
+        self.cB = np.array([cP.copy() for cP in self.cS])
         
-        self._progress_log()
+        self._finalize_iteration()
         
 
     def _run(self):
+        """Main loop of BA method.
+
+        Returns
+        -------
+        optimum: Bat
+            Best solution found during the BA optimization.
+        """
+        
         self._check_params()
+        
         err_msg = self._init_method()
-        if err_msg:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
 
         if 'pulse_rate' in self.params:
             r_ = self.params['pulse_rate']
@@ -136,7 +185,7 @@ class BA(Optimizer):
             cP.A = alpha*cP.A
             cP.r = r_ *(1 - np.exp(-gamma*self.it))
         
-        for self.it in range(1, self.iterations + 1):
+        while True:
             
             A_avg = np.mean(np.array([self.cS[p].A for p in range(len(self.cS))]))
             
@@ -151,7 +200,7 @@ class BA(Optimizer):
                 if np.random.uniform() > cP.r:
                     cP.X = self.best.X + 0.05*np.abs(self.lb - self.ub)*np.random.normal(size=self.dimensions)*A_avg
                                        
-                cP.X = np.clip(cP.X, self.lb, self.ub)
+                cP.clip()
                
             #Evaluate swarm
             for p, cP in enumerate(self.cS):
@@ -165,13 +214,10 @@ class BA(Optimizer):
             if err_msg:
                 break
             
-            self._progress_log()
-            # Check stopping conditions
-            if self._stopping_criteria():
+            if self._finalize_iteration():
                 break
             
-        if not err_msg:        
-            return self.best
-        else:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
+        
+        return self.best
