@@ -1,18 +1,28 @@
 # -*- coding: utf-8 -*-
+"""DIFFERENTIAL EVOLUTION METHOD"""
 
 import numpy as np
 from ._optimizer import Optimizer, CandidateState 
 from scipy.stats import cauchy
 
-"""
-R. Tanabe and A. S. Fukunaga, “Improving the search performance of SHADE using 
-linear population size reduction”, in Proceedings of the 2014 IEEE Congress 
-on EvolutionaryComputation (CEC), pp. 1658–1665, Beijing, China, July 2014.
-"""
-
 
 class Solution(CandidateState):
-    """DE solution class"""
+    """DE solution class.
+    
+    Parameters
+    ----------
+    V : ndarray
+        Mutant vector.  
+    CR : float
+        DE control parameter CR.
+    F : float
+        DE control parameter F.
+         
+    Returns
+    -------
+    solution : Solution
+        Solution instance.
+    """
     
     def __init__(self, optimizer: Optimizer):
         CandidateState.__init__(self, optimizer)
@@ -24,21 +34,46 @@ class Solution(CandidateState):
 
 
 class DE(Optimizer):
-    """Differential Evolution class"""
+    """Differential Evolution method class.
+    
+    Reference: R. Tanabe and A. S. Fukunaga. Improving the search performance of SHADE using 
+    linear population size reduction”, in Proceedings of the 2014 IEEE Congress on 
+    Evolutionary Computation (CEC), pp. 1658–1665, Beijing, China, July 2014.
+    
+    Parameters
+    ----------
+    variant : str
+        Name of the DE variant (``SHADE`` or ``LSHADE``). Default: ``SHADE``.
+    params : dict
+        A dictionary of DE parameters.
+    """
 
     def __init__(self):
-        """Initialization"""
         Optimizer.__init__(self)
         #super(PSO, self).__init__() # ugly version of the above
 
-        self.method = 'SHADE'
+        self.variant = 'SHADE'
         self.params = {}
 
     def _check_params(self):
+        """Private method which performs some DE-specific parameter checks
+        and prepares the parameters to be validated by Optimizer._check_params.
+
+        Returns
+        -------
+        None
+            Nothing
+        """
+        
         defined_params = list(self.params.keys())
         mandatory_params, optional_params = [], []
+        
+        if 'initial_population_size' in self.params:
+            self.params['initial_population_size'] = int(self.params['initial_population_size'])
+        if 'historical_memory_size' in self.params:
+            self.params['historical_memory_size'] = int(self.params['historical_memory_size'])
 
-        if self.method == 'SHADE':
+        if self.variant == 'SHADE':
             mandatory_params = 'initial_population_size external_archive_size_factor historical_memory_size p_mutation'.split()
             if 'initial_population_size' not in self.params:
                 self.params['initial_population_size'] = self.dimensions * 18
@@ -56,7 +91,7 @@ class DE(Optimizer):
             if 'rank_enabled' not in self.params:
                 self.params['rank_enabled'] = False # Rank-based variant off by default
                 defined_params += 'rank_enabled'.split()  
-        elif self.method == 'LSHADE':
+        elif self.variant == 'LSHADE':
             mandatory_params = 'initial_population_size external_archive_size_factor historical_memory_size p_mutation'.split()
             if 'initial_population_size' not in self.params:
                 self.params['initial_population_size'] = self.dimensions * 18
@@ -75,7 +110,7 @@ class DE(Optimizer):
                 self.params['rank_enabled'] = False # Rank-based variant off by default
                 defined_params += 'rank_enabled'.split()
         else:
-            assert False, f'Unknown method! {self.method}'
+            assert False, f'Unknown variant! {self.variant}'
         
         if self.constraints > 0:
             assert False, 'DE does not support constraints'
@@ -92,12 +127,17 @@ class DE(Optimizer):
         Optimizer._check_params(self, mandatory_params, optional_params, defined_params)
         
     def _init_method(self):
-        
+        """Private method for initializing the DE optimizer instance.
+        Initializes and evaluates the population.
+
+        Returns
+        -------
+        err_msg : str or None
+            Error message (if any).
+        """
+
+        self._evaluate_initial_candidates()
         err_msg = None
-        
-        # Bounds for position
-        self.lb = np.array(self.lb)
-        self.ub = np.array(self.ub)
 
         # Generate a population
         self.Pop = np.array([Solution(self) for c in \
@@ -115,36 +155,46 @@ class DE(Optimizer):
         self.M_F = np.full(self.params['historical_memory_size'], 0.5)
 
         # Generate initial positions
+        n0 = 0 if self._cs0 is None else self._cs0.size
         for i in range(self.params['initial_population_size']):
             
             # Random position
             self.Pop[i].X = np.random.uniform(self.lb, self.ub)
             
             # Using specified particles initial positions
-            if self.X0 is not None:
-                if i < np.shape(self.X0)[0]:
-                    self.Pop[i].X = self.X0[i]
+            if i < n0:
+                self.Pop[i] = self._cs0[i].copy()
 
         # Evaluate
-        self.collective_evaluation(self.Pop)
-        # if all candidates are NaNs       
+        if n0 < self.params['initial_population_size']:
+            err_msg = self.collective_evaluation(self.Pop[n0:])
+
+        # if all candidates are NaNs
         if np.isnan([p.f for p in self.Pop]).all():
             err_msg = 'ALL CANDIDATES FAILED TO EVALUATE.'
         if err_msg:
             return err_msg
         
-        self._progress_log()
+        self._finalize_iteration()
         
     def _run(self):
+        """Main loop of DE method.
+
+        Returns
+        -------
+        optimum: Solution
+            Best solution found during the DE optimization.
+        """
+        
         self._check_params()
+        
         err_msg = self._init_method()
-        if err_msg:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
         
         k = 0 # memory index
 
-        for self.it in range(1, self.iterations + 1):
+        while True:
             
             S_CR = np.empty([0])
             S_F = np.empty([0])
@@ -247,22 +297,17 @@ class DE(Optimizer):
                     k = 0
                     
             # Linear Population Size Reduction (LPSR)
-            if self.method == 'LSHADE':
+            if self.variant == 'LSHADE':
                 N_init = self.params['initial_population_size']
                 N_new = round((4 - N_init) * self._progress_factor() + N_init)
                 if N_new < np.size(self.Pop):
                     self.Pop = np.sort(self.Pop)[:N_new]
                     self.Trials = self.Trials[:N_new]          
                 
-            # Update history
-            self._progress_log()
-            
-            # Check stopping conditions
-            if self._stopping_criteria():
+            if self._finalize_iteration():
                 break
 
-        if not err_msg:        
-            return self.best
-        else:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
+        
+        return self.best
