@@ -4,22 +4,21 @@ from ._optimizer import Optimizer, CandidateState
 import numpy as np
 
 class NelderMead(Optimizer):
-    """Squirrel Search Algorithm class"""
+    """Nelder Mead method class."""
 
     def __init__(self):
-        """Initialization"""
         Optimizer.__init__(self)
 
         self.X = None
         self.X0 = None
-        self.method = 'Vanilla'
+        self.variant = 'Vanilla'
         self.params = {}
         self.params['init_step'] = 0.4
         self.params['alpha'] = 1.0
         self.params['gamma'] = 2.0
         self.params['rho'] = 0.5
         self.params['sigma'] = 0.5
-        self.iterations = 100
+        self.max_iterations = 100
 
 
     def _check_params(self):
@@ -27,7 +26,7 @@ class NelderMead(Optimizer):
         defined_params = list(self.params.keys())
         mandatory_params, optional_params = [], []
 
-        if self.method == 'Vanilla':
+        if self.variant == 'Vanilla':
             mandatory_params = ''.split()
             optional_params = 'init_step alpha gamma rho sigma'.split()
 
@@ -38,14 +37,14 @@ class NelderMead(Optimizer):
 
         for param in defined_params:
             if param not in mandatory_params and param not in optional_params:
-                print(f'Warning: Excessive parameter {param}')
+                self.log(f'Warning: Excessive parameter {param}')
 
         Optimizer._check_params(self, mandatory_params, optional_params, defined_params)
 
     def _init_method(self):
-        # Bounds for position and velocity
-        self.lb = np.array(self.lb)
-        self.ub = np.array(self.ub)
+
+        self._evaluate_initial_candidates()
+        err_msg = None
 
         # Generate set of points
         self.cS = np.array([CandidateState(self) for _ in range(self.dimensions + 1)], \
@@ -68,6 +67,7 @@ class NelderMead(Optimizer):
             dx = np.zeros([self.dimensions])
             dx[p - 1] = self.params['init_step']
             self.cS[p].X = self.cS[0].X + dx * (self.ub - self.lb)
+            self.cS[p].X = np.clip(self.cS[p].X, self.lb, self.ub)
 
             # Using specified particles initial positions
             if self.X0 is not None:
@@ -82,7 +82,7 @@ class NelderMead(Optimizer):
         self._check_params()
         self._init_method()
 
-        for self.it in range(self.iterations):
+        while True:
             self.cS = np.sort(self.cS)
             # print(self.it, np.min(self.cS).f)
             #print(i, self.cS[0].f, self.cS[-1].f)
@@ -169,30 +169,46 @@ class NelderMead(Optimizer):
                 # self.cS[p].evaluate()
             self.collective_evaluation(self.cS[1:])
 
+            if self._finalize_iteration():
+                break
 
         return self.best
 
 
 class MSGS(Optimizer):
-    """Multi Scale Grid Search class"""
+    """Multi Scale Grid Search class.
+
+    Parameters
+    ----------
+    _x : CandidateState
+        CandidateState for current point.
+    _y : ndarray of int
+        The search grid indices for current point.
+    """
 
     def __init__(self):
-        """Initialization"""
         Optimizer.__init__(self)
-        self.X0 = None
+        self.X0 = 1
         self.X = None
-        self.method = 'Vanilla'
-        self.params = {}
-        self.params['n'] = 10
-        self.params['xtol'] = 1e-4
-        self.iterations = 100
+        self.variant = 'Vanilla'
+        self.params = {'n': 10, 'xtol': 1e-6}
 
     def _check_params(self):
+        """Private method which prepares the parameters to be validated by Optimizer._check_params.
+        
+        Returns
+        -------
+        None
+            Nothing
+        """
 
         defined_params = list(self.params.keys())
         mandatory_params, optional_params = [], []
+        
+        if 'n' in self.params:
+            self.params['n'] = int(self.params['n'])
 
-        if self.method == 'Vanilla':
+        if self.variant == 'Vanilla':
             mandatory_params = ''.split()
             optional_params = 'n xtol'.split()
 
@@ -203,73 +219,162 @@ class MSGS(Optimizer):
 
         for param in defined_params:
             if param not in mandatory_params and param not in optional_params:
-                print(f'Warning: Excessive parameter {param}')
+                self.log(f'Warning: Excessive parameter {param}')
 
         Optimizer._check_params(self, mandatory_params, optional_params, defined_params)
 
-
     def _init_method(self):
-        # Bounds for position and velocity
-        self.lb = np.array(self.lb)
-        self.ub = np.array(self.ub)
+        """Private method for initializing the MSGS optimizer instance.
+        Evaluates given initial candidates, selects starting point, constructs and deforms 
+        (if needed) initial grid, initializes grid search history.
+
+        Returns
+        -------
+        None
+            Nothing
+        """
+        
+        self._evaluate_initial_candidates()
 
         self._n = self.params['n']
         self._I = np.arange(-self._n, self._n + 1)
+        self._x = self._cs0[0].copy()
 
-        self._x = CandidateState(self)
-        if self.X0 is not None:
-            self._x.X = self.X0[0, :]
-        else:
-            self._x.X = np.random.uniform(self.lb, self.ub)
-
-        self._reinit_grid(0.5 * (self.lb + self.ub), 1, 'Initial grid')
+        # The initial mesh always fills entire domain and initial point is snapped tzo the grid
+        # self._center = 0.5 * (self.lb + self.ub)
+        self._center = self._x.X
+        self._X_map = [self._I for i in range(self.dimensions)]
+        self._reinit_grid(self._center, 1, 'Initial grid')
         self._x = self._y_to_x(self._y)
 
-        self.collective_evaluation([self._x])
+        # self.collective_evaluation([self._x])
         self._H.append(list(self._y))
-        self.log(f'Initial point: {self._x.X}')
+        self.log(f'Initial point: [' + ', '.join([f'{x}' for x in self._x.X]) + ']')
+
+        self._finalize_iteration()
 
     def _reinit_grid(self, center, scale, grid_action):
+        """(Re)creates the search grid according to given center of the grid and grid scale. Grid action is only
+        informative argument and it does not influence the generation of the search grid. If current point
+        (``MSGS._x``) is not on the grid, only the closest discrete values are adjusted to the current point resulting
+        with locally deformed search grid.
+
+        Parameters
+        ----------
+        center : ndarray
+            The center of the search grid. Grid is initially constructed using ``n`` divisions before and after the
+            center (2n + 1 divisions). If needed, due to bounds, center is automatically shifted and adjusted in order to ensure the search
+            mesh is inside the bounds.
+
+        scale: int
+            The scale of the search grid. The initial grid is generated using ``scale=1`` which covers entire domain.
+            The discretization step for each optimization variable is defined as
+            ``dx = (self.ub - self.lb) / (2 * self._n * scale)``.
+
+        grid_action: string
+            Information about the reason why the search grid reinitialization is being conducted. Possible values are:
+            ``none``, ``scale_up``, ``scale_down`` or ``shift``.
+        """
         # self.log(f'Init grid @{self.best.X if self.best else "None"} x{self._scale}')
         dx = (self.ub - self.lb) / (2 * self._n * scale)
 
-        new_center = center.copy()
+        # Snap center to dx grid
+        new_center = np.round(center.copy() / dx) * dx
         for i in range(self.dimensions):
             xi = new_center[i] + self._I * dx[i]
-            # self.log(f'{xi=}')
             dil = int(np.sum(xi < self.lb[i]))
             new_center[i] += dx[i] * dil
             diu = int(np.sum(xi > self.ub[i]))
             new_center[i] -= dx[i] * diu
-            if dil > 0 or diu > 0:
-                pass
-                # Bounds_reached
 
-        self._X_map = [center[i] + self._I * dx[i] for i in range(self.dimensions)]
+            if dil > 0 and diu > 0:
+                assert False, 'HELLO!'
+
+            xi = new_center[i] + self._I * dx[i]
+            jc = np.argmin(np.abs(xi - center[i]))
+            if jc == -self._n:
+                jc += 1
+            if jc == self._n:
+                jc -= 1
+
+            self._X_map[i] = new_center[i] + self._I * dx[i]
+
+            d = (center[i] - new_center[i]) / dx[i]
+
+            self._X_map[i][jc] = new_center[i] + d * dx[i]
+            # self.log(f'{self._X_map[i]=}')
+
+        # self._X_map = [new_center[i] + self._I * dx[i] for i in range(self.dimensions)]
         self._y = self._x_to_y(self._x)
         self._H = [list(self._y)]
 
         self.log(f'Grid reinitialization {scale=}, {grid_action=}')
-        if np.all(np.isclose(center, new_center, self.params['xtol'] * 1e-3)):
+        if np.all(np.isclose(self._center, new_center, (self.ub - self.lb) * self.params['xtol'] * 1e-3)):
             # Center unchanged
+            # self.log(f'c: {self._center}')
+            # self.log(f'c: {new_center}')
             return False
         else:
             # Center changed
             return True
 
-    def _x_to_y(self, x):
+    def _x_to_y(self, x: CandidateState):
+        """Converts the candidate defined with coordinates in the search space to indices of candidates location on the
+        search grid.
+
+        Parameters
+        ----------
+        x : CandidateState
+            The candidate state you want to transform.
+
+        Returns
+        -------
+        y : ndarray of int
+            The indices of the search grid where point x is located.
+        """
+
         y = np.zeros(self.dimensions, dtype=int)
         for i in range(self.dimensions):
             y[i] = np.argmin(np.abs(x.X[i] - self._X_map[i])) - self._n
         return y
 
-    def _y_to_x(self, y):
+    def _y_to_x(self, y: np.ndarray):
+        """Converts the indices of the search grid to a candidate.
+
+        Parameters
+        ----------
+        y : ndarray of int
+            The indices of the search grid where point x is located.
+
+        Returns
+        -------
+        x : CandidateState
+            The candidate state you want to transform.
+        """
+
         x = CandidateState(self)
         for i in range(self.dimensions):
             x.X[i] = self._X_map[i][y[i] + self._n]
         return x
 
     def _find_dir(self, y, dy=None):
+        """Finds a search direction based on finite difference approximation of a gradiend using points on the search
+        grid. It utilizes a 2-point or 3-point gradient calculation, if existing direction ``dy`` is given or not,
+        respectively.
+
+        Parameters
+        ----------
+        y : ndarray of int
+            The coordinates of current point defined using indices of the search grid.
+        dy : ndarray of int
+            A step utilized in previous step of the method defined as an array of possible directions ``[-1, 0, 1]``
+            for each dimension.
+
+        Returns
+        -------
+        new_direction : ndarray of int
+            New direction in terms of steps (one of ``[-1, 0, 1]``) on the search grid for each dimension.
+        """
 
         # self.log(f'{y=}')
         # self.log(f'{self._x.X=}')
@@ -303,7 +408,8 @@ class MSGS(Optimizer):
                 eval_map[2 * i + 1] = cnt
                 cnt += 1
 
-        self.collective_evaluation(C)
+        if C.size > 0:
+            self.collective_evaluation(C)
 
         for i in range(self.dimensions):
             c0 = self._x
@@ -331,6 +437,14 @@ class MSGS(Optimizer):
         return new_dy
 
     def _run(self):
+        """Main loop of MSGS method.
+
+        Returns
+        -------
+        optimum: CandidateState
+            Best solution found during the MSGS optimization.
+        """
+        
         self._check_params()
         self._init_method()
 
@@ -342,20 +456,22 @@ class MSGS(Optimizer):
         action = 'find_dir_3p'
         shift_count = 0
 
-        for self.it in range(self.iterations):
+        while True:
 
             # self.log(f'{grid_action=}, {action=}')
             self._update_history()  # Adding history record allows iterations without evaluation
 
             if grid_action == 'shift':
-                if shift_count > 2 and scale >= 2:
-                    scale = scale / 2
+                if shift_count >= 2 and scale >= 2:
+                    scale = int(scale / 2)
                     shift_count = 0
+                    self.log('scale_down (3rd consecutive shift)')
 
                 if self._reinit_grid(self._x.X, scale, grid_action):
                     # Genter is changed
-                    #action = 'move'
                     grid_action = 'none'
+                    shift_count += 1
+                    self.log(f'shift ({shift_count=})')
                 else:
                     scale = scale * 8
                     self._reinit_grid(self.best.X, scale, 'scale_up (shift failed)')
@@ -367,6 +483,7 @@ class MSGS(Optimizer):
                 self._reinit_grid(self._x.X, scale, grid_action)
                 action = 'find_dir_3p'
                 grid_action = 'none'
+                shift_count = 0
 
             elif grid_action == 'scale_down':
                 if scale >= 2:
@@ -374,13 +491,13 @@ class MSGS(Optimizer):
                 self._reinit_grid(self._x.X, scale)
                 action = 'find_dir_3p'
                 grid_action = 'none'
+                shift_count = 0
 
             elif grid_action == 'none':
                 pass
             else:
                 assert False, f'Unknown {grid_action=}'
 
-
             if action == 'find_dir_3p':
                 # Finding a new direction based on the full gradient calculation
                 dy = self._find_dir(self._y)
@@ -393,35 +510,45 @@ class MSGS(Optimizer):
                 k += 1
 
             elif action == 'check_dir':
-                # Moving in dy direction
-                new_y = self._y + dy
-                # Trimming bounds
-                for i in range(d):
-                    new_y[new_y < -self._n] = - self._n
-                    new_y[new_y > self._n] = self._n
-                dy = new_y - self._y
+                # self.log(f'{dy=}')
 
-                if np.sum(np.abs(dy)) == 0:
+                if np.all(dy == 0):
                     # None of x_next's neighbouring points are better
                     if self._x == self.best:
                         # x_next is the best, hence convergence is finished
                         action = 'converged'
                     else:
                         action = 'move_to_best'
+
                 else:
-                    action = 'move'
-                    bound_y = np.abs(self._y) == self._n
-                    if np.any(bound_y):
-                        new_x = self._y_to_x(self._y).X
+                    # Moving in dy direction
+                    new_y = self._y + dy
 
-                        bound_x = np.logical_or(np.isclose(new_x, self.lb),
-                                                np.isclose(new_x, self.ub))
+                    # Check bounds
+                    # outside_y = new_y != np.clip(new_y, -self._n, self._n, dtype=int)
+                    outside_y = np.logical_or(new_y < -self._n, new_y > self._n)
+                    if not np.any(outside_y):
+                        # New point is inside the mesh
+                        # dy = new_y - self._y
+                        action = 'move'
+                    else:
+                        # self.log(f'bounds problem {dy=}')
+                        # New point is outside the mesh
+                        # Calculate x coordinates of the new point
+                        dx = (self.ub - self.lb) / (2 * self._n * scale)
+                        new_x = self._y_to_x(self._y).X + dy * dx
 
-                        if np.sum(bound_y) == np.sum(bound_x):
-                            # Reached lb/ub bounds, no rescaling needed
-                            pass
+                        outside_x = np.logical_or(new_x < self.lb, new_x > self.ub)
+
+                        # Correct the step
+                        dy[outside_x] = 0
+                        # self.log(f'corrected step {dy=}')
+                        if np.any(np.abs(dy) > 0):
+                            action = 'move'
                         else:
-                            # action = 'none'
+                            action == 'converged'
+
+                        if np.any(np.logical_and(np.abs(dy) > 0, np.logical_and(outside_y, np.logical_not(outside_x)))):
                             grid_action = 'shift'
 
             elif action == 'move_to_best':
@@ -434,6 +561,7 @@ class MSGS(Optimizer):
             elif action == 'move':
                 # Moving in dy direction
                 self._y = self._y + dy
+                self._y = np.clip(self._y, -self._n, self._n, dtype=int)
 
                 if list(self._y) in self._H:
                     action = 'move_to_best'
@@ -465,11 +593,7 @@ class MSGS(Optimizer):
             else:
                 assert False, f'Unknown action {action=}'
 
-            self._progress_log()
-
-            # Check stopping conditions
-            if self._stopping_criteria():
-                self.log('Stop')
+            if self._finalize_iteration():
                 break
 
         return self.best
