@@ -1,44 +1,68 @@
 # -*- coding: utf-8 -*-
+"""Electromagnetic Field Optimization method"""
 
 import numpy as np
 from ._optimizer import Optimizer, CandidateState 
 from scipy.constants import golden_ratio as phi
 
-"""
-Electromagnetic Field Optimization
-
-Abedinpourshotorban, H., Shamsuddin, S.M., Beheshti, Z., & Jawawi, D.N. (2016). 
-Electromagnetic field optimization: a physics-inspired metaheuristic 
-optimization algorithm. Swarm and Evolutionary Computation, 26, 8-22.
-
-Notes: Due to evaluating only one particle per iteration, the method:
-* needs many iterations to be efficient (much more than other methods)
-* is effectively not parallelized in collective_evaluation 
-  (hence parallelization is not allowed, in order to avoid user confusion)
-"""
 
 class emParticle(CandidateState):
-    """EFO Particle class"""
+    """EFO emParticle class. An emParticle is a member of EFO population.
+         
+    Returns
+    -------
+    emp : emParticle
+        emParticle instance.
+    """
     
     def __init__(self, optimizer: Optimizer):
         CandidateState.__init__(self, optimizer)
 
 
 class EFO(Optimizer):
-    """Electromagnetic Field Optimization class"""
+    """Electromagnetic Field Optimization method class.
+    
+    Reference: Abedinpourshotorban, H., Shamsuddin, S.M., Beheshti, Z., & Jawawi, D.N. (2016). 
+    Electromagnetic field optimization: a physics-inspired metaheuristic 
+    optimization algorithm. Swarm and Evolutionary Computation, 26, 8-22.
+    
+    Due to evaluating only one particle per iteration, this method:
+    
+    - needs many iterations to be efficient (much more than other methods)
+    
+    - is effectively not parallelized in **collective_evaluation** (hence parallel evaluation is not allowed, in order to avoid user confusion)
+    
+    Parameters
+    ----------
+    variant : str
+        Name of the EFO variant. Default: ``Vanilla``.
+    params : dict
+        A dictionary of EFO parameters.
+    """
 
     def __init__(self):
-        """Initialization"""
         Optimizer.__init__(self)
 
-        self.method = 'Vanilla'
+        self.variant = 'Vanilla'
         self.params = {}
 
     def _check_params(self):
+        """Private method which performs some EFO-specific parameter checks
+        and prepares the parameters to be validated by Optimizer._check_params.
+
+        Returns
+        -------
+        None
+            Nothing
+        """
+        
         defined_params = list(self.params.keys())
         mandatory_params, optional_params = [], []
+        
+        if 'population_size' in self.params:
+            self.params['population_size'] = int(self.params['population_size'])
 
-        if self.method == 'Vanilla':
+        if self.variant == 'Vanilla':
             mandatory_params = 'population_size R_rate Ps_rate P_field N_field'.split()
             if 'population_size' not in self.params:
                 self.params['population_size'] = 10 * self.dimensions
@@ -56,16 +80,26 @@ class EFO(Optimizer):
                 self.params['N_field'] = 0.45 # recommended [0.4, 0.5]
                 defined_params += 'N_field'.split()
         else:
-            assert False, f'Unknown method! {self.method}'
+            assert False, f'Unknown variant! {self.variant}'
             
         if self.number_of_processes > 1:
-            assert False, 'EFO does not support parallelization'
+            self.log('Warning: EFO does not support parallelization. Defaulting to number_of_processes=1.')
+            self.number_of_processes = 1
 
         Optimizer._check_params(self, mandatory_params, optional_params, defined_params)
       
                     
     def _init_method(self):
-        
+        """Private method for initializing the EFO optimizer instance.
+        Initializes and evaluates the swarm.
+
+        Returns
+        -------
+        err_msg : str or None
+            Error message (if any).
+        """
+
+        self._evaluate_initial_candidates()
         err_msg = None
         
         # Generate a population
@@ -75,20 +109,21 @@ class EFO(Optimizer):
         self.cNew = emParticle(self)
         
         # Initialize
+        n0 = 0 if self._cs0 is None else self._cs0.size
         for i, p in enumerate(self.cS):
             
             # Random position
-            p.X = np.random.uniform(self.lb, self.ub)
+            self.cS[i].X = np.random.uniform(self.lb, self.ub)
             
             # Using specified particles initial positions
-            if self.X0 is not None:
-                if i < np.shape(self.X0)[0]:
-                    p.X = self.X0[i]
+            if i < n0:
+                self.cS[i] = self._cs0[i].copy()
                     
         self.cNew.X = np.random.uniform(self.ub, self.lb)
         
-        # Evaluate        
-        err_msg = self.collective_evaluation(self.cS)
+        # Evaluate
+        if n0 < self.cS.size:
+            err_msg = self.collective_evaluation(self.cS[n0:])
         self.cS = np.sort(self.cS)
         # if all candidates are NaNs       
         if np.isnan([cP.f for cP in self.cS]).all():
@@ -96,15 +131,23 @@ class EFO(Optimizer):
         if err_msg:
             return err_msg
         
-        self._progress_log()
+        self._finalize_iteration()
         
 
     def _run(self):
+        """Main loop of EFO method.
+
+        Returns
+        -------
+        optimum: emParticle
+            Best solution found during the EFO optimization.
+        """
+        
         self._check_params()
+        
         err_msg = self._init_method()
-        if err_msg:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
         
         if 'population_size' in self.params:
             N_emp = self.params['population_size']
@@ -119,7 +162,7 @@ class EFO(Optimizer):
                         
         RI = 0
         
-        for self.it in range(1, self.iterations + 1):
+        while True:
                           
             for d in range(self.dimensions):
                 force = np.random.uniform(0,1)
@@ -148,26 +191,17 @@ class EFO(Optimizer):
             
             err_msg = self.collective_evaluation(np.array([self.cNew], dtype=emParticle))
             if err_msg:
-                self.log(err_msg)
                 break
             
-            # insert cNew if better than last in list
-            p_last = self.cS[-1]                   
-            if self.cNew < p_last:
-                p_last.X = np.copy(self.cNew.X)
-                p_last.f = self.cNew.f
-                p_last.O = np.copy(self.cNew.O)
-                p_last.C = np.copy(self.cNew.C)   
+            # insert cNew if better than last in list    
+            if self.cNew < self.cS[-1]:
+                self.cS[-1] = self.cNew.copy()
             self.cS = np.sort(self.cS) 
                     
-            self._progress_log()
-            
-            # Check stopping conditions
-            if self._stopping_criteria():
+            if self._finalize_iteration():
                 break
             
-        if not err_msg:        
-            return self.best
-        else:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
+        
+        return self.best
