@@ -1,26 +1,48 @@
 # -*- coding: utf-8 -*-
+""" FIREWORKS ALGORITHM method"""
 
 import numpy as np
 from ._optimizer import Optimizer, CandidateState 
 
 
 class FWA(Optimizer):
-    """Firework Algorithm class"""
+    """Fireworks Algorithm method class.
+    
+    Parameters
+    ----------
+    variant : str
+        Name of the FWA variant (``Vanilla`` or ``Rank``). Default: ``Rank``.
+    params : dict
+        A dictionary of FWA parameters.
+    """
 
     def __init__(self):
-        """Initialization"""
         Optimizer.__init__(self)
 
-        # self.X = None # seems to be obsolete
-        self.cX = None
-        self.method = 'Rank'
+        self.variant = 'Rank'
         self.params = {}
 
     def _check_params(self):
+        """Private method which performs some FWA-specific parameter checks
+        and prepares the parameters to be validated by Optimizer._check_params.
+
+        Returns
+        -------
+        None
+            Nothing
+        """
+        
         defined_params = list(self.params.keys())
         mandatory_params, optional_params = [], []
+        
+        if 'n' in self.params:
+            self.params['n'] = int(self.params['n'])
+        if 'm1' in self.params:
+            self.params['m1'] = int(self.params['m1'])
+        if 'm2' in self.params:
+            self.params['m2'] = int(self.params['m2'])
 
-        if self.method == 'Vanilla':
+        if self.variant == 'Vanilla':
             mandatory_params = 'n m1 m2'.split()
             if 'n' not in self.params:
                 self.params['n'] = self.dimensions
@@ -32,7 +54,7 @@ class FWA(Optimizer):
                 self.params['m2'] = self.dimensions // 2
                 defined_params += 'm2'.split()                
             optional_params = ''.split()
-        elif self.method == 'Rank':
+        elif self.variant == 'Rank':
             mandatory_params = 'n m1 m2'.split()
             if 'n' not in self.params:
                 self.params['n'] = self.dimensions
@@ -45,32 +67,43 @@ class FWA(Optimizer):
                 defined_params += 'm2'.split()
             optional_params = ''.split()
         else:
-            assert False, f'Unknown method \'{self.method}\''
+            assert False, f'Unknown variant \'{self.variant}\''
         
         if self.constraints > 0:
-            assert self.method == 'Rank', f'Method \'{self.method}\' does not support constraints! Use \'Rank\' method instead'
+            assert self.variant == 'Rank', f'Variant \'{self.variant}\' does not support constraints! Use \'Rank\' variant instead'
 
         Optimizer._check_params(self, mandatory_params, optional_params, defined_params)
 
     def _init_method(self):
-        
+        """Private method for initializing the FWA optimizer instance.
+        Initializes and evaluates the swarm.
+
+        Returns
+        -------
+        err_msg : str or None
+            Error message (if any).
+        """
+
+        self._evaluate_initial_candidates()
         err_msg = None
 
-        self.cX = np.array([CandidateState(self) for p in range(self.params['n'])], dtype=CandidateState)
+        self.cX = np.array([CandidateState(self) for p in range(self.params['n'])])
         
         # Generate initial positions
+        n0 = 0 if self._cs0 is None else self._cs0.size
         for p in range(self.params['n']):
             
             # Random position
             self.cX[p].X = np.random.uniform(self.lb, self.ub, self.dimensions)
             
             # Using specified initial positions
-            if self.X0 is not None:
-                if p < np.shape(self.X0)[0]:
-                    self.cX[p].X = self.X0[p]
+            if p < n0:
+                self.cX[p] = self._cs0[p].copy()
 
         # Evaluate all
-        err_msg = self.collective_evaluation(self.cX)               
+        if n0 < self.params['n']:
+            err_msg = self.collective_evaluation(self.cX[n0:])
+
         # if all candidates are NaNs       
         if np.isnan([p.f for p in self.cX]).all():
             err_msg = 'ALL CANDIDATES FAILED TO EVALUATE.'
@@ -80,27 +113,30 @@ class FWA(Optimizer):
         # Sort
         self.cX = np.sort(self.cX)
         
-        # Initialize the results - this is probably handled by progress_log
-        # self.results = OptimizationResults(self)
-        # self.results.cHistory = [np.min(self.cX).copy()]
+        self._finalize_iteration()
+
+
+    def _run(self):
+        """Main loop of FWA method.
+
+        Returns
+        -------
+        optimum: CandidateState
+            Best solution found during the FWA optimization.
+        """
         
-        self._progress_log()
-
-
-    def _run(self):#, seed=None):
-        #Optimizer.run(self, seed=seed)
         self._check_params()
+        
         err_msg = self._init_method()
-        if err_msg:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
         
         n = self.params['n']
 
-        for self.it in range(1, self.iterations + 1):
+        while True:
 
-            explosion_sparks = self.explosion()
-            mutation_sparks = self.gaussian_mutation()
+            explosion_sparks = self._explosion()
+            mutation_sparks = self._gaussian_mutation()
 
             #self.__mapping_rule(sparks, self.lb, self.ub, self.dimensions)
             for cS in (explosion_sparks + mutation_sparks):
@@ -118,22 +154,24 @@ class FWA(Optimizer):
 
             self.cX = np.sort(self.cX)[:n]
 
-            # Update history
-            #self.results.cHistory.append(np.min(self.cX).copy()) # superseded by progress_log()
-            self._progress_log()
-
-            # Check stopping conditions
-            if self._stopping_criteria():
+            if self._finalize_iteration():
                 break
         
-        if not err_msg:        
-            return self.best
-        else:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
+        
+        return self.best
 
 
-    def explosion(self):
+    def _explosion(self):
+        """Private method for computing explosion sparks.
+
+        Returns
+        -------
+        explosion_sparks : list of CandidateState
+            FWA explosion sparks.
+        """
+        
         eps=0.001
         amp=10
         a=0.01
@@ -148,10 +186,10 @@ class FWA(Optimizer):
             cFw = self.cX[p].copy()
             #print(cFw.X)
             
-            if self.method == 'Vanilla':
+            if self.variant == 'Vanilla':
                 # Number of sparks
                 n1 = self.params['m1'] * (fmax - cFw.f + eps) / np.sum(fmax - F + eps)
-                n1 = self.min_max_round(n1, self.params['m1'] * a, self.params['m2'] * b)
+                n1 = self._min_max_round(n1, self.params['m1'] * a, self.params['m2'] * b)
                 
                 # Amplitude
                 A = amp * (cFw.f - fmin + eps) /  (np.sum(F - fmin) + eps)
@@ -162,11 +200,11 @@ class FWA(Optimizer):
                             cFw.X[k] += np.random.uniform(-A, A)
                     explosion_sparks.append(cFw.copy())
                 
-            if self.method == 'Rank':
+            if self.variant == 'Rank':
                 
                 # Number of sparks
                 #vn1 = self.params['m1'] * (fmax - cFw.f + eps) / np.sum(fmax - F + eps)
-                #vn1 = self.min_max_round(vn1, self.params['m1'] * a, self.params['m2'] * b)
+                #vn1 = self._min_max_round(vn1, self.params['m1'] * a, self.params['m2'] * b)
                 
                 n1 = self.params['m1'] * (self.params['n'] - p)**1 / np.sum(np.arange(self.params['n']+1)**1)
                 n1 = np.random.choice([int(np.floor(n1)), int(np.ceil(n1))])
@@ -210,25 +248,14 @@ class FWA(Optimizer):
         #input(' > Press return to continue.')
         return explosion_sparks
     
-    """
-    def __explosion_operator(self, sparks, fw, function,
-                             dimension, m, eps, amp, Ymin, Ymax, a, b):
-        
-        sparks_num = self.__round(m * (Ymax - function(fw) + eps) /
-                                  (sum([Ymax - function(fwk) for fwk in self.X]) + eps), m, a, b)
-        print(sparks_num)
+    def _gaussian_mutation(self):
+        """Private method for computing mutation sparks.
 
-        amplitude = amp * (function(fw) - Ymax + eps) / \
-            (sum([function(fwk) - Ymax for fwk in self.X]) + eps)
-
-        for j in range(int(sparks_num)):
-            sparks.append(np.array(fw))
-            for k in range(dimension):
-                if (np.random.choice([True, False])):
-                    sparks[-1][k] += np.random.uniform(-amplitude, amplitude)
-    """
-    
-    def gaussian_mutation(self):
+        Returns
+        -------
+        mutation_sparks : list of CandidateState
+            FWA mutation sparks.
+        """
         
         mutation_sparks = []
         for j in range(self.params['m2']):
@@ -242,7 +269,44 @@ class FWA(Optimizer):
         #print('mut sparks:', np.sort([p.f for p in mutation_sparks]))
         #print('mut sparks:', len(mutation_sparks))
         return mutation_sparks
-            
+    
+    def _min_max_round(self, s, smin, smax):
+        """Private method for calculating round of min of max of input parameters.
+
+        Parameters
+        ----------
+        s : float
+            Preliminary population size.
+        smin : float
+            Preliminary population size.
+        smax : float
+            Preliminary population size.
+             
+        Returns
+        -------
+        min_max_round : int
+            Round of min of max of input parameters.
+        """
+        
+        return int(np.round(np.min([np.max([s, smin]), smax])))
+    
+    """
+    def __explosion_operator(self, sparks, fw, function,
+                             dimension, m, eps, amp, Ymin, Ymax, a, b):
+        
+        sparks_num = self.__round(m * (Ymax - function(fw) + eps) /
+                                  (sum([Ymax - function(fwk) for fwk in self.X]) + eps), m, a, b)
+        #print(sparks_num)
+
+        amplitude = amp * (function(fw) - Ymax + eps) / \
+            (sum([function(fwk) - Ymax for fwk in self.X]) + eps)
+
+        for j in range(int(sparks_num)):
+            sparks.append(np.array(fw))
+            for k in range(dimension):
+                if (np.random.choice([True, False])):
+                    sparks[-1][k] += np.random.uniform(-amplitude, amplitude)
+                    
     def __mapping_rule(self, sparks, lb, ub, dimension):
         for i in range(len(sparks)):
             for j in range(dimension):
@@ -254,9 +318,6 @@ class FWA(Optimizer):
         self.collective_evaluation(sparks)
         self.cX = np.append(self.cX, sparks)
 
-    def min_max_round(self, s, smin, smax):
-        return int(np.round(np.min([np.max([s, smin]), smax])))
-
     def __round(self, s, m, a, b):
         if (s < a * m):
             return round(a * m)
@@ -264,3 +325,4 @@ class FWA(Optimizer):
             return round(b * m)
         else:
             return round(s)
+    """
