@@ -21,13 +21,16 @@ class Solution(CandidateState):
     """DE solution class"""
     
     def __init__(self, optimizer: Optimizer):
-        CandidateState.__init__(self, optimizer)
-        #super(Particle, self).__init__(optimizer) # ugly version of the above
-        
+        CandidateState.__init__(self, optimizer)        
         self.CR = None
         self.F = None
         self.V = np.zeros([optimizer.dimensions]) * np.nan # mutant vector
 
+class Bee(CandidateState):
+    """ABC Bee class"""
+    def __init__(self, optimizer: Optimizer):
+        CandidateState.__init__(self, optimizer)
+
 
 class MMO(Optimizer):
     """Mutualistic Multi-Optimization class"""
@@ -37,10 +40,7 @@ class MMO(Optimizer):
         Optimizer.__init__(self)
 
         self.params = {}        
-        self.method = {}
-        
-        # PSO
-        # no special parameters
+        self.methods = {}
         
         # FWA  
         self.cX = None
@@ -50,15 +50,19 @@ class MMO(Optimizer):
         defined_params = list(self.params.keys())
         mandatory_params, optional_params = [], []
         
-        if not self.method:
-            self.method = {'PSO': 'Vanilla', 
+        if not self.methods:
+            self.methods = {'PSO': 'Vanilla', 
                            'FWA': 'Rank'}
-        assert len(self.method) >= 2, \
-            'optimizer.method should provide at least 2 optimization methods'
+            
+        assert len(self.methods) >= 2, \
+            'optimizer.methods should provide at least 2 optimization methods'
         
-        if 'PSO' in self.method:        
+        if 'PSO' in self.methods:      
+            
+            if 'swarm_size' in self.params:
+                self.params['swarm_size'] = int(self.params['swarm_size'])
     
-            if self.method['PSO'] == 'Vanilla':
+            if self.methods['PSO'] == 'Vanilla':
                 mandatory_params += 'swarm_size inertia cognitive_rate social_rate'.split()
                 if 'swarm_size' not in self.params:
                     self.params['swarm_size'] = self.dimensions
@@ -73,7 +77,7 @@ class MMO(Optimizer):
                     self.params['social_rate'] = 1.0
                     defined_params += 'social_rate'.split()
                 optional_params += 'akb_model akb_fun_start akb_fun_stop'.split()
-            elif self.method['PSO'] == 'TVAC':
+            elif self.methods['PSO'] == 'TVAC':
                 mandatory_params += 'swarm_size inertia'.split()
                 if 'swarm_size' not in self.params:
                     self.params['swarm_size'] = self.dimensions
@@ -83,11 +87,18 @@ class MMO(Optimizer):
                     defined_params += 'inertia'.split()
                 optional_params += 'akb_model akb_fun_start akb_fun_stop'.split()
             else:
-                assert False, f"Unknown PSO method! {self.method['PSO']}"
+                assert False, f"Unknown PSO variant! {self.methods['PSO']}"
         
-        if 'FWA' in self.method:
+        if 'FWA' in self.methods:
             
-            if self.method['FWA'] == 'Vanilla':
+            if 'n' in self.params:
+                self.params['n'] = int(self.params['n'])
+            if 'm1' in self.params:
+                self.params['m1'] = int(self.params['m1'])
+            if 'm2' in self.params:
+                self.params['m2'] = int(self.params['m2'])
+            
+            if self.methods['FWA'] == 'Vanilla':
                 mandatory_params += 'n m1 m2'.split()
                 if 'n' not in self.params:
                     self.params['n'] = self.dimensions
@@ -99,7 +110,7 @@ class MMO(Optimizer):
                     self.params['m2'] = self.dimensions // 2
                     defined_params += 'm2'.split()    
                 optional_params += ''.split()
-            elif self.method['FWA'] == 'Rank':
+            elif self.methods['FWA'] == 'Rank':
                 mandatory_params += 'n m1 m2'.split()
                 if 'n' not in self.params:
                     self.params['n'] = self.dimensions
@@ -112,14 +123,17 @@ class MMO(Optimizer):
                     defined_params += 'm2'.split()
                 optional_params += ''.split()
             else:
-                assert False, f"Unknown FWA method \'{self.method['FWA']}\'"
+                assert False, f"Unknown FWA variant \'{self.methods['FWA']}\'"
             
             if self.constraints > 0:
-                assert self.method == 'Rank', f"FWA Method \'{self.method['FWA']}\' does not support constraints! Use \'Rank\' method instead"
+                assert self.methods == 'Rank', f"FWA variant \'{self.methods['FWA']}\' does not support constraints! Use \'Rank\' method instead"
                 
-        if 'DE' in self.method:
+        if 'DE' in self.methods:
             
-            if self.method['DE'] == 'SHADE':
+            if 'initial_population_size' in self.params:
+                self.params['initial_population_size'] = int(self.params['initial_population_size'])
+            
+            if self.methods['DE'] == 'SHADE':
                 mandatory_params += 'initial_population_size external_archive_size_factor historical_memory_size p_mutation'.split()
                 if 'initial_population_size' not in self.params:
                     self.params['initial_population_size'] = self.dimensions * 18
@@ -134,7 +148,7 @@ class MMO(Optimizer):
                     self.params['p_mutation'] = 0.11
                     defined_params += 'p_mutation'.split()    
                 optional_params = ''.split()
-            elif self.method['DE'] == 'LSHADE':
+            elif self.methods['DE'] == 'LSHADE':
                 mandatory_params += 'initial_population_size external_archive_size_factor historical_memory_size p_mutation'.split()
                 if 'initial_population_size' not in self.params:
                     self.params['initial_population_size'] = self.dimensions * 18
@@ -150,12 +164,32 @@ class MMO(Optimizer):
                     defined_params += 'p_mutation'.split()  
                 optional_params = ''.split()
             else:
-                assert False, f"Unknown DE method! {self.method['DE']}"
-            
+                assert False, f"Unknown DE variant! {self.methods['DE']}"
+
             if self.constraints > 0:
                 assert False, 'DE does not support constraints'
 
-        if 'PSO' in self.method:
+        if 'ABC' in self.methods:
+            
+            if self.methods['ABC'] == 'Vanilla':
+                mandatory_params += 'bees trial_limit'.split()
+                
+                if 'bees' in self.params:
+                    self.params['bees'] = int(self.params['bees'])
+                else:
+                    self.params['bees'] = self.dimensions
+                defined_params += 'bees'.split()
+                
+                if 'trial_limit' in self.params:
+                    self.params['trial_limit'] = int(self.params['trial_limit'])            
+                else:
+                    self.params['trial_limit'] = int((self.params['bees']*self.dimensions)/2) # Karaboga and Gorkemli 2014 - "A quick artificial bee colony (qabc) algorithm and its performance on optimization problems"
+                    defined_params += 'trial_limit'.split()
+            
+            else:
+                assert False, f"Unknown ABC variant! {self.methods['ABC']}"
+
+        if 'PSO' in self.methods:
             
             """ Anakatabatic Inertia a.k.a. Polynomial PFIDI """
             if self.params['inertia'] == 'anakatabatic':
@@ -173,26 +207,27 @@ class MMO(Optimizer):
                             w_start = [-0.86, 0.24, -1.10, 0.75, 0.72]
                             w_stop = [-0.81, -0.35, -0.26, 0.64, 0.60]
                             splinetype = 'linear'
-                            if self.method['PSO'] != 'Vanilla':
-                                print('Warning: akb_model \'FlyingStork\' was designed for Vanilla PSO')                    
+                            if self.methods['PSO'] != 'Vanilla':
+                                self.log('Warning: akb_model \'FlyingStork\' was designed for Vanilla PSO')                    
                         elif self.params['akb_model'] == 'MessyTie':
                             w_start = [-0.62, 0.18, 0.65, 0.32, 0.77]
                             w_stop = [0.36, 0.73, -0.62, 0.40, 1.09]
                             splinetype = 'linear'
-                            if self.method['PSO'] != 'Vanilla':
-                                print('Warning: akb_model \'MessyTie\' was designed for Vanilla PSO')   
+                            if self.methods['PSO'] != 'Vanilla':
+                                self.log('Warning: akb_model \'MessyTie\' was designed for Vanilla PSO')   
                         elif self.params['akb_model'] == 'RightwardPeaks':
                             w_start = [-1.79, -0.33, 2.00, -0.67, 1.30]
                             w_stop = [-0.91, -0.88, -0.84, 0.67, -0.36]
                             splinetype = 'linear'
-                            if self.method['PSO'] != 'TVAC':
-                                print('Warning: akb_model \'RightwardPeaks\' was designed for TVAC PSO')
+                            if self.methods['PSO'] != 'TVAC':
+                                self.log('Warning: akb_model \'RightwardPeaks\' was designed for TVAC PSO')
                         elif self.params['akb_model'] == 'OrigamiSnake':
                             w_start = [-1.36, 2.00, 1.00, -0.60, 1.22]
                             w_stop = [0.30, 1.03, -0.21, 0.40, 0.06]
                             splinetype = 'linear'
-                            if self.method['PSO'] != 'TVAC':
-                                print('Warning: akb_model \'OrigamiSnake\' was designed for TVAC PSO')                      # code shared for all w-list-based named akb_models
+                            if self.methods['PSO'] != 'TVAC':
+                                self.log('Warning: akb_model \'OrigamiSnake\' was designed for TVAC PSO')
+                        # code shared for all w-list-based named akb_models
                         Th = np.linspace(np.pi/4, 5*np.pi/4, 5)
                         self.params['akb_fun_start'] = \
                                             interp1d(Th, w_start, kind=splinetype)
@@ -200,7 +235,7 @@ class MMO(Optimizer):
                                             interp1d(Th, w_stop, kind=splinetype) 
                     else:
                         if self.params['akb_model'] != 'Languid':
-                            print('Warning: Unknown akb_model. Defaulting to \'Languid\'')
+                            self.log('Warning: Unknown akb_model. Defaulting to \'Languid\'')
                             self.params['akb_model'] = 'Languid'
                     if self.params['akb_model'] == 'Languid':
                         def akb_fun_languid(Th):
@@ -223,7 +258,7 @@ class MMO(Optimizer):
         self.lb = np.array(self.lb)
         self.ub = np.array(self.ub)
                
-        if 'PSO' in self.method:
+        if 'PSO' in self.methods:
             
             """ Initialize PSO """
             # Bounds for velocity
@@ -261,7 +296,7 @@ class MMO(Optimizer):
                 return err_msg
             
             # Use initial particles as best ones
-            self.cB = np.array([cP.copy() for cP in self.cS], dtype=CandidateState)
+            self.cB = np.array([cP.copy() for cP in self.cS])
             
             # Update the overall best
             self.p_best = np.argmin(self.cB)
@@ -271,15 +306,15 @@ class MMO(Optimizer):
             # self.progress_log()
             
             self.BI = np.zeros(self.params['swarm_size'], dtype=int)
-            self.TOPO = np.zeros([self.params['swarm_size'], self.params['swarm_size']], dtype=np.bool)
+            self.TOPO = np.zeros([self.params['swarm_size'], self.params['swarm_size']], dtype=np.bool_)
     
             self.reinitialize_topology()
             self.find_neighbourhood_best()
         
-        if 'FWA' in self.method:
+        if 'FWA' in self.methods:
             
             """ Initialize FWA """
-            self.cX = np.array([CandidateState(self) for p in range(self.params['n'])], dtype=CandidateState)
+            self.cX = np.array([CandidateState(self) for p in range(self.params['n'])])
             
             # Generate initial positions
             for p in range(self.params['n']):
@@ -293,7 +328,7 @@ class MMO(Optimizer):
                         self.cX[p].X = self.X0[p]
     
             # Evaluate all
-            self.collective_evaluation(self.cX)
+            err_msg = self.collective_evaluation(self.cX)
             # if all candidates are NaNs       
             if np.isnan([p.f for p in self.cX]).all():
                 err_msg = 'ALL CANDIDATES FAILED TO EVALUATE.'
@@ -303,7 +338,7 @@ class MMO(Optimizer):
             # Sort
             self.cX = np.sort(self.cX)
         
-        if 'DE' in self.method:
+        if 'DE' in self.methods:
             
             """ Initialize DE """
             # Generate a population
@@ -322,27 +357,65 @@ class MMO(Optimizer):
             self.M_F = np.full(self.params['historical_memory_size'], 0.5)
     
             # Generate initial positions
+            n0 = 0 if self._cs0 is None else self._cs0.size
             for i in range(self.params['initial_population_size']):
                 
                 # Random position
                 self.Pop[i].X = np.random.uniform(self.lb, self.ub)
                 
                 # Using specified particles initial positions
-                if self.X0:
-                    if i < np.shape(self.X0)[0]:
-                        self.Pop[i].X = self.X0[i]
+                if i < n0:
+                    self.Pop[i] = self._cs0[i].copy()
 
             # Evaluate
-            self.collective_evaluation(self.Pop)
+            err_msg = self.collective_evaluation(self.Pop)
+            
             # if all candidates are NaNs       
             if np.isnan([p.f for p in self.Pop]).all():
                 err_msg = 'ALL CANDIDATES FAILED TO EVALUATE.'
             if err_msg:
                 return err_msg
         
+        if 'ABC' in self.methods:
+            
+            """ Initialize ABC """
+            # Generate a swarm
+            self.ABC_cS = np.array([Bee(self) for c in range(self.params['bees'])], dtype=Bee)
+            self.ABC_cS_k = np.array([Bee(self) for c in range(self.params['bees'])], dtype=Bee)
+
+            # Generate initial trial and probability vectors
+
+            self.trials = np.zeros([self.params['bees']],dtype=np.int_)
+            self.probability = np.zeros([self.params['bees']])
+            
+            n0 = 0 if self._cs0 is None else self._cs0.size
+            for p in range(self.params['bees']):
+                
+                # Random position
+                self.ABC_cS[p].X =  np.random.uniform(self.lb, self.ub)
+
+                # Using specified particles initial positions
+                if p < n0:
+                    self.ABC_cS[p] = self._cs0[p].copy()
+                
+            # Evaluate
+            if n0 < self.params['bees']:
+                err_msg = self.collective_evaluation(self.ABC_cS[n0:])
+            # err_msg = self.collective_evaluation(self.ABC_cS)
+            # err_msg = self.collective_evaluation(self.ABC_cS_k)
+            #self.ABC_cS_k = np.copy(self.ABC_cS)
+
+            # if all candidates are NaNs       
+            if np.isnan([cP.f for cP in self.ABC_cS]).all():
+                err_msg = 'ALL CANDIDATES FAILED TO EVALUATE.'
+            
+            if err_msg:
+                return err_msg
+            
+            self.ABC_cB = np.array([cP.copy() for cP in self.ABC_cS])
+        
         """ After all methods are initialized... """
-        # Update history
-        self._progress_log()
+        self._finalize_iteration()
  
        
     """ PSO functions """
@@ -380,7 +453,7 @@ class MMO(Optimizer):
             cFw = self.cX[p].copy()
             #print(cFw.X)
             
-            if self.method['FWA'] == 'Vanilla':
+            if self.methods['FWA'] == 'Vanilla':
                 # Number of sparks
                 n1 = self.params['m1'] * (fmax - cFw.f + eps) / np.sum(fmax - F + eps)
                 n1 = self.min_max_round(n1, self.params['m1'] * a, self.params['m2'] * b)
@@ -394,7 +467,7 @@ class MMO(Optimizer):
                             cFw.X[k] += np.random.uniform(-A, A)
                     explosion_sparks.append(cFw.copy())
                 
-            if self.method['FWA'] == 'Rank':
+            if self.methods['FWA'] == 'Rank':
                 
                 # Number of sparks
                 #vn1 = self.params['m1'] * (fmax - cFw.f + eps) / np.sum(fmax - F + eps)
@@ -471,21 +544,69 @@ class MMO(Optimizer):
         else:
             return round(s)
     
+    
     """ DE functions """
     
     # no special DE functions
     
+    
+    """ ABC functions """
+    
+    def _employed_bees_phase(self):
+        
+        err_msg = None
+        
+        for p, cP in enumerate(self.ABC_cS_k):
+        
+            k = np.random.randint(0,self.params['bees'])
+            while k == p:
+                k = np.random.randint(0,self.params['bees'])
+     
+            d = np.random.randint(0,self.dimensions)
+            phi = np.random.uniform(-1,1)
+            
+            for i in range(self.dimensions):
+                if i != d: self.ABC_cS_k[p].X[i] = self.ABC_cS[p].X[i]
+     
+            self.ABC_cS_k[p].X[d] = self.ABC_cS[p].X[d] + phi*(self.ABC_cS[p].X[d] - self.ABC_cS[k].X[d])
+            
+            cP.clip()
+                
+        for cP in self.ABC_cS:
+            cP.clip()
+
+        err_msg = self.collective_evaluation(self.ABC_cS_k)
+        
+        for p, cP in enumerate(self.ABC_cS_k):
+            if self.ABC_cS_k[p] < self.ABC_cS[p]:
+                self.ABC_cS[p] = self.ABC_cS_k[p].copy()
+                self.trials[p] = 0
+            else:
+                self.trials[p] = self.trials[p] + 1
+    
+        if err_msg:
+            return err_msg
+    
+    
+    def _fitness_calc(self):
+        temp_fitness = np.array([])
+            
+        for p, cP in enumerate(self.ABC_cS):
+            temp_fitness = np.append(temp_fitness, self.ABC_cS[p].f)
+        
+        return temp_fitness
+    
 
     """ MMO run function """
     def _run(self):
         self._check_params()
+
         err_msg = self._init_method()
-        if err_msg:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
 
         """ Prepare PSO """
-        if 'PSO' in self.method:   
+        if 'PSO' in self.methods:   
             if 'inertia' in self.params.keys():
                 w = self.params['inertia']
             if 'cognitive_rate' in self.params.keys():
@@ -494,24 +615,29 @@ class MMO(Optimizer):
                 c2 = self.params['social_rate']
 
         """ Prepare FWA """
-        if 'FWA' in self.method:
+        if 'FWA' in self.methods:
             n = self.params['n']
         
         """ Prepare DE """
         # no special preparement needed
         
+        """ Prepare ABC """
+        if 'ABC' in self.methods:
+            if 'trial_limit' in self.params:
+                trial_limit = self.params['trial_limit']
+        
         """ Main run loop """
-        for self.it in range(1, self.iterations + 1):
+        while True:
             
             """ PSO iteration """
-            if 'PSO' in self.method:
+            if 'PSO' in self.methods:
                 R1 = np.random.uniform(0, 1, [self.params['swarm_size'], self.dimensions])
                 R2 = np.random.uniform(0, 1, [self.params['swarm_size'], self.dimensions])
     
                 if self.params['inertia'] == 'LDIW':
                     w = 1.0 - (1.0 - 0.4) * self._progress_factor()
     
-                if self.method['PSO'] == 'TVAC':
+                if self.methods['PSO'] == 'TVAC':
                     c1 = 2.5 - (2.5 - 0.5) * self._progress_factor()
                     c2 = 0.5 + (2.5 - 0.5) * self._progress_factor()
     
@@ -541,7 +667,7 @@ class MMO(Optimizer):
                     cP.X = cP.X + cP.V        
                     
                     # Correct position to the bounds
-                    cP.X = np.clip(cP.X, self.lb, self.ub)       
+                    cP.clip()    
                     
                 # Get old fitness
                 f_old = np.array([cP.f for cP in self.cS])
@@ -563,13 +689,6 @@ class MMO(Optimizer):
                 # Update the overall best
                 self.p_best = np.argmin(self.cB)
                 
-                # Update history
-                self._progress_log()
-                
-                # Check stopping conditions
-                if self._stopping_criteria():
-                    break
-                
                 # Update swarm topology
                 if np.max(self.dF) >= 0.0:
                     self.reinitialize_topology()
@@ -578,7 +697,7 @@ class MMO(Optimizer):
                 self.find_neighbourhood_best()
             
             """ FWA iteration """
-            if 'FWA' in self.method:
+            if 'FWA' in self.methods:
                 explosion_sparks = self.explosion()
                 mutation_sparks = self.gaussian_mutation()
     
@@ -597,16 +716,9 @@ class MMO(Optimizer):
                     break
     
                 self.cX = np.sort(self.cX)[:n]
-    
-                # Update history
-                self._progress_log()
-    
-                # Check stopping conditions
-                if self._stopping_criteria():
-                    break
             
             """ DE iteration """
-            if 'DE' in self.method:
+            if 'DE' in self.methods:
                 K = 0 # memory index
                     
                 S_CR = np.empty([0])
@@ -681,67 +793,127 @@ class MMO(Optimizer):
                         K = 0
                         
                 # Linear Population Size Reduction (LPSR)
-                if self.method == 'LSHADE':
+                if self.methods == 'LSHADE':
                     N_init = self.params['initial_population_size']
                     N_new = round((4 - N_init) * self._progress_factor() + N_init)
                     if N_new < np.size(self.Pop):
                         self.Pop = np.sort(self.Pop)[:N_new]
                         self.Trials = self.Trials[:N_new]          
-                    
-                # Update history
-                self._progress_log()
+            
+            """ ABC iteration """
+            if 'ABC' in self.methods:
                 
-                # Check stopping conditions
-                if self._stopping_criteria():
-                    break
+                """employed bees phase"""
+                err_msg = self._employed_bees_phase()
+            
+                
+                """probability update"""
+                for p, cP in enumerate(self.ABC_cS):
+                    # self.probability[p] = 0.1 + 0.9 * (self.ABC_cS[p].f/np.max(temp_fitness))
+                    self.probability[p] = self.ABC_cS[p].f / np.sum(self._fitness_calc())
+                    
+                """onlooker bee phase"""
+                i = 0
+                t = 0
+                while t < self.params['bees']:
+                    
+                    if np.random.uniform(0,1) < self.probability[i]:
+                        
+                        t = t + 1
+                        err_msg = self._employed_bees_phase()
+       
+                    i = (i + 1)%(self.params['bees'] - 1)
+
+                
+                """scout bee phase"""
+                for p, cP in enumerate(self.ABC_cS):
+                    if self.trials[p] > trial_limit:
+                        self.ABC_cS[p].X = np.random.uniform(self.lb, self.ub)
+                        self.trials[p] = 0
+                
+                if err_msg:
+                    break    
             
             """ Methods cooperating """
             bests = {}
-            if 'PSO' in self.method:
+            if 'PSO' in self.methods:
                 bests['PSO'] = self.cB[self.p_best].f
-            if 'FWA' in self.method:
+            if 'FWA' in self.methods:
                 bests['FWA'] = np.min(self.cX).f
-            if 'DE' in self.method:
-                bests['DE'] = np.min(self.Pop).f 
+            if 'DE' in self.methods:
+                bests['DE'] = np.min(self.Pop).f
+            if 'ABC' in self.methods:
+                bests['ABC'] = np.min(self.ABC_cS).f
             
             if min(bests, key=bests.get) == 'PSO': # PSO best is THE BEST
-                if 'FWA' in self.method:
+                if 'FWA' in self.methods:
                     FWAworst = np.max(self.cX)
                     FWAworst.X = np.copy(self.best.X) 
                     FWAworst.f = self.best.f
-                if 'DE' in self.method:
+                if 'DE' in self.methods:
                     DEworst = np.max(self.Pop)
                     DEworst.X = np.copy(self.best.X) 
                     DEworst.f = self.best.f
+                if 'ABC' in self.methods:
+                    ABCworst = np.max(self.ABC_cS)
+                    ABCworst.X = np.copy(self.best.X) 
+                    ABCworst.f = self.best.f
                 #print('PSO updated others')
             elif min(bests, key=bests.get) == 'FWA': # FWA best is THE BEST
-                if 'PSO' in self.method:
+                if 'PSO' in self.methods:
                     PSOworst = np.max(self.cB)
                     PSOworst.X = np.copy(self.best.X) 
                     PSOworst.f = self.best.f 
                     PSOworst.dF = 0.0
                     self.find_neighbourhood_best()
-                if 'DE' in self.method:
+                if 'DE' in self.methods:
                     DEworst = np.max(self.Pop)
                     DEworst.X = np.copy(self.best.X) 
                     DEworst.f = self.best.f
+                if 'ABC' in self.methods:
+                    ABCworst = np.max(self.ABC_cS)
+                    ABCworst.X = np.copy(self.best.X) 
+                    ABCworst.f = self.best.f
                 #print('FWA updated others')
-            elif min(bests, key=bests.get) == 'DE': # FWA best is THE BEST
-                if 'PSO' in self.method:
+            elif min(bests, key=bests.get) == 'DE': # DE best is THE BEST
+                if 'PSO' in self.methods:
                     PSOworst = np.max(self.cB)
                     PSOworst.X = np.copy(self.best.X) 
                     PSOworst.f = self.best.f 
                     PSOworst.dF = 0.0
                     self.find_neighbourhood_best()
-                if 'FWA' in self.method:
+                if 'FWA' in self.methods:
                     FWAworst = np.max(self.cX)
                     FWAworst.X = np.copy(self.best.X) 
                     FWAworst.f = self.best.f
+                if 'ABC' in self.methods:
+                    ABCworst = np.max(self.ABC_cS)
+                    ABCworst.X = np.copy(self.best.X) 
+                    ABCworst.f = self.best.f
                 #print('DE updated others')
-       
-        if not err_msg:        
-            return self.best
-        else:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+            elif min(bests, key=bests.get) == 'ABC': # ABC best is THE BEST
+                if 'PSO' in self.methods:
+                    PSOworst = np.max(self.cB)
+                    PSOworst.X = np.copy(self.best.X) 
+                    PSOworst.f = self.best.f 
+                    PSOworst.dF = 0.0
+                    self.find_neighbourhood_best()
+                if 'FWA' in self.methods:
+                    FWAworst = np.max(self.cX)
+                    FWAworst.X = np.copy(self.best.X) 
+                    FWAworst.f = self.best.f
+                if 'DE' in self.methods:
+                    DEworst = np.max(self.Pop)
+                    DEworst.X = np.copy(self.best.X) 
+                    DEworst.f = self.best.f
+                #print('ABC updated others')
+            
+            """ Wrap up iteration """
+            if self._finalize_iteration():
+                break
+ 
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
+        
+        return self.best
     
\ No newline at end of file
