@@ -4,38 +4,57 @@ import numpy as np
 from ._optimizer import Optimizer, CandidateState 
 from scipy.special import gamma
 
-"""
-Jain, M., Singh, V., & Rani, A. (2019). A novel nature-inspired algorithm for 
-optimization: Squirrel search algorithm. Swarm and evolutionary computation, 
-44, 148-175.
-"""
-
 
 class FlyingSquirrel(CandidateState):
-    """SSA agent class"""
+    """SSA agent class.
+    
+    Returns
+    -------
+    FS : FlyingSquirrel
+        FlyingSquirrel instance.        
+    """
     
     def __init__(self, optimizer: Optimizer):
         CandidateState.__init__(self, optimizer)
 
 
 class SSA(Optimizer):
-    """Squirrel Search Algorithm class"""
+    """Squirrel Search Algorithm method class.
+    
+    Reference: Jain, M., Singh, V., & Rani, A. (2019). A novel nature-inspired algorithm 
+    for optimization: Squirrel search algorithm. Swarm and evolutionary computation, 
+    44, 148-175.
+    
+    Parameters
+    ----------
+    variant : str
+        Name of the SSA variant. Default: ``Vanilla``.
+    params : dict
+        A dictionary of SSA parameters.
+    """
 
-    def __init__(self):
-        """Initialization"""
-        
+    def __init__(self):        
         Optimizer.__init__(self)
 
-        self.X = None # is this necessary?
-        self.method = 'Vanilla'
+        self.variant = 'Vanilla'
         self.params = {}
 
     def _check_params(self):
+        """Private method which prepares the parameters to be validated by Optimizer._check_params.
 
+        Returns
+        -------
+        None
+            Nothing
+        """
+        
         defined_params = list(self.params.keys())
         mandatory_params, optional_params = [], []
+        
+        if 'swarm_size' in self.params:
+            self.params['swarm_size'] = int(self.params['swarm_size'])
 
-        if self.method == 'Vanilla':
+        if self.variant == 'Vanilla':
             mandatory_params = 'swarm_size acorn_tree_attraction'.split()
             # the following params are better left at default
             mandatory_params += 'predator_presence_probability gliding_constant \
@@ -58,11 +77,19 @@ class SSA(Optimizer):
                 defined_params += 'gliding_distance_limits'.split()
             optional_params = ''.split()
         else:
-            assert False, f'Unknown method! {self.method}'
+            assert False, f'Unknown variant! {self.variant}'
             
         Optimizer._check_params(self, mandatory_params, optional_params, defined_params)
         
     def _init_method(self):
+        """Private method for initializing the SSA optimizer instance.
+        Initializes and evaluates the swarm.
+
+        Returns
+        -------
+        None
+            Nothing
+        """
         
         err_msg = None
 
@@ -71,35 +98,40 @@ class SSA(Optimizer):
                             dtype=FlyingSquirrel)
         
         # Generate initial positions
+        n0 = 0 if self._cs0 is None else self._cs0.size
         for p in range(self.params['swarm_size']):            
             # Random position
             self.cS[p].X = np.random.uniform(self.lb, self.ub)           
             # Using specified particles initial positions
-            if self.X0 is not None:
-                if p < np.shape(self.X0)[0]:
-                    self.cS[p].X = self.X0[p]                                          
+            if p < n0:
+                self.cS[p] = self._cs0[p].copy()
         
         # Evaluate
-        self.collective_evaluation(self.cS)
+        if n0 < self.params['swarm_size']:
+            err_msg = self.collective_evaluation(self.cS[n0:])
+
         # if all candidates are NaNs       
         if np.isnan([cP.f for cP in self.cS]).all():
             err_msg = 'ALL CANDIDATES FAILED TO EVALUATE.'
         if err_msg:
             return err_msg
         
-        # Find the overall best, i.e. FSht (hickory nut tree)
-        #self.best = np.sort(self.cS)[0] # already done in collective_evaluation
-            
-        # Update history
-        #self.results.cHistory = [self.best] # superseded by progress_log()
-        self._progress_log()
+        self._finalize_iteration()
         
     def _run(self):
+        """Main loop of SSA method.
+
+        Returns
+        -------
+        optimum: FlyingSquirrel
+            Best solution found during the PSO optimization.
+        """
+        
         self._check_params()
+        
         err_msg = self._init_method()
-        if err_msg:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
       
         # Load params
         if 'acorn_tree_attraction' in self.params:
@@ -121,7 +153,7 @@ class SSA(Optimizer):
                      gamma((1 + beta) / 2) * beta * 2**((beta - 1)/2)) **(1 / beta)
             return 0.01 * (ra * sigma) / (np.abs(rb)**(1 / beta))
 
-        for self.it in range(1, self.iterations + 1):
+        while True:
             
             # Categorizing FS's
             FSht = np.sort(self.cS)[0] # best FS (hickory nut trees)
@@ -182,30 +214,19 @@ class SSA(Optimizer):
                 for fs in FSnt:
                     fs.X = self.lb + Levy() * (self.ub - self.lb)
             
-            for cP in self.cS:
-                # Correct position to the bounds
-                cP.X = np.clip(cP.X, self.lb, self.ub)       
+            # Correct position to the bounds
+            for cP in self.cS:               
+                cP.clip()       
                 
             # Evaluate swarm
             err_msg = self.collective_evaluation(self.cS)
             if err_msg:
                 break
-            
-            # Update the overall best
-            # self.best = np.min(self.cS)  
-            # This is a bug? Updating global best with iterational best
-            # self.best is automatically updated in collective_evaluation
              
-            # Update history
-            #self.results.cHistory.append(self.best.copy()) # superseded by progress_log()
-            self._progress_log()
-                
-            # Check stopping conditions
-            if self._stopping_criteria():
+            if self._finalize_iteration():
                 break
             
-        if not err_msg:        
-            return self.best
-        else:
-            print('Error: ' + err_msg + ' OPTIMIZATION ABORTED\n')
-            return
\ No newline at end of file
+        assert not err_msg, \
+            f'Error: {err_msg} OPTIMIZATION ABORTED'
+        
+        return self.best
\ No newline at end of file
